import yaml
import os
import click
import zipfile
import psutil
import urllib

def getinit_data(initfiles,parameters):
    '''
    get initial data from both a list of files and a list of 'pname=pvalue'
    strings as they are passed in the command line <pvalue> is assumed to be a
    YAML parsable string.
    '''
    initdata = {}
    for initfile in initfiles:
        initdata.update(**yaml.load(open(initfile)))

    for x in parameters:
        key,value = x.split('=')
        initdata[key]=yaml.load(value)
    return initdata

def prepare_workdir_from_archive(workdir,inputarchive):
    if os.path.exists(workdir):
        raise click.exceptions.ClickException(click.style("workdirectory exists and input archive give. Can't have both", fg = 'red'))
    inputdata = '{}/inputs'.format(workdir)
    os.makedirs(inputdata)
    localzipfile = '{}/inputarchive.zip'.format(workdir)
    urllib.urlretrieve(inputarchive,localzipfile)
    with zipfile.ZipFile(localzipfile) as zf:
        zf.extractall(path = inputdata)
    os.remove(localzipfile)

def setupbackend_fromstring(backend, name = 'backendname'):
    if backend.startswith('multiproc'):
        import backends.packtivitybackend as pb
        nparallel  = backend.split(':')[1]
        if nparallel == 'auto':
            nparallel = psutil.cpu_count()
        else:
            nparallel = int(nparallel)
        backend = pb.PacktivityMultiProcBackend(nparallel)
    elif backend == 'celery':
        import backends.celeryapp
        import backends.packtivity_celery as pc
        backend = pc.PacktivityCeleryBackend(backends.celeryapp.app)
    elif backend == 'foreground':
        import backends.packtivitybackend as pb
        backend = pb.PacktivityForegroundBackend()
    elif backend == 'jira':
        import backends.jira as jb
        backend = jb.JiraBackend('workflow request - {}'.format(name),'some description')
    else:
        raise NotImplementedError('backend config {} not known'.format(backend))

    return backend



#!/usr/bin/env python
import click
import os
import steering_api
import logging
import capschemas
import clihelpers

log = logging.getLogger(__name__)

@click.command()
@click.option('-v','--verbosity', default = 'INFO')
@click.option('-i','--loginterval', default = 30)
@click.option('-u','--updateinterval', default = 0.02)
@click.option('-c','--schemasource', default = capschemas.schemadir)
@click.option('-b','--backend', default = 'multiproc:auto')
@click.option('--interactive/--not-interactive', default = False)
@click.option('--validate/--no-validate', default = True)
@click.option('--visualize/--no-visualize', default = True) #v below this we should only have options/arg available also in yadage-manual
@click.option('-t','--toplevel', default = os.getcwd())
@click.option('-a','--inputarchive', default = None)
@click.option('--parameter', '-p', multiple=True)
@click.argument('workdir')
@click.argument('workflow')
@click.argument('initfiles', nargs = -1)
def main(workdir,
         workflow,
         initfiles,
         toplevel,
         verbosity,
         loginterval,
         updateinterval,
         schemasource,
         backend,
         interactive,
         parameter,
         validate,
         visualize,
         inputarchive):
    logging.basicConfig(level = getattr(logging,verbosity))

    if inputarchive:
        clihelpers.prepare_workdir_from_archive(workdir,inputarchive)

    initdata = clihelpers.getinit_data(initfiles,parameter)
    backend  = clihelpers.setupbackend_fromstring(backend)

    rc =  steering_api.run_workflow(
            workdir,
            workflow,
            initdata,
            toplevel,
            updateinterval,
            loginterval,
            validate = validate,
            doviz = visualize,
            schemadir = schemasource,
            backend = backend,
            user_interaction = interactive
    )
    if rc:
        exc = click.exceptions.ClickException(click.style("Workflow failed", fg = 'red'))
        exc.exit_code = rc
        raise exc

if __name__ == '__main__':
    main()



import hashlib
import uuid
import json
import os
from backends.staticbackend import StaticProxy

def set_backend(dag,backend,proxymaker):
    '''
    sets backend and proxies for each node in the DAG.
    proxymaker is a 1-ary function that takes the node object and
    returns a suitable result proxy
    '''
    for nodename in dag.nodes():
        n = dag.getNode(nodename)
        n.backend = backend
        n.resultproxy = proxymaker(n)

def set_static_backend(dag,backend):
    '''
    sets the backend for the case of a static backend
    Proxies are set using the node identifier
    '''
    set_backend(
        dag,
        backend,
        proxymaker = lambda n: StaticProxy(n.identifier)
    )

def json_hash(jsonable):
    return hashlib.sha1(json.dumps(jsonable, cls = WithJsonRefEncoder, sort_keys = True)).hexdigest()

def get_id_fromjson(jsonobject, method = 'uuid'):
    method = os.environ.get('YADAGE_ID_METHOD',method)
    if method == 'uuid':
        return str(uuid.uuid1())
    elif method == 'jsonhash':
        return json_hash(jsonobject)
    else:
        raise NotImplementedError('unkown id generation method {}'.format(method))


def get_obj_id(obj_with_json_method, method = 'uuid'):
    return get_id_fromjson(obj_with_json_method.json(),method)


import jsonref
class WithJsonRefEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, jsonref.JsonRef):
            return {k:v for k,v in obj.iteritems()}
        else:
            super(WithJsonRefEncoder,self).default(obj)



import shlex
import pydotplus
import jsonpointer
import logging
import jq
import subprocess

log = logging.getLogger(__name__)

def fillscope(cluster,workflow,scope = ''):
    # scopecluster = stagecluster = pydotplus.graphviz.Cluster(graph_name = '_'.join(stagescopeprts),
    scopecluster = pydotplus.graphviz.Cluster(
                        graph_name = scope.replace('/',''),
                        label = ''.join(['[{}]'.format(p) for p in scope.split('/')[1:]]),
                        style = 'solid',
                        color = 'blue')
    cluster.add_subgraph(scopecluster)
    scopeptr = jsonpointer.JsonPointer(scope)
    scoped = scopeptr.resolve(workflow.stepsbystage)
    for stage,elements in scoped.iteritems():
        stagescopeprts = scopeptr.parts+[stage]
        stagecluster = pydotplus.graphviz.Cluster(
            graph_name = '_'.join(stagescopeprts),
            label = stage,
            labeljust = 'l',
            color = 'grey',
            style = 'dashed')
        scopecluster.add_subgraph(stagecluster)
        for i,element in enumerate(elements):
            if '_nodeid' in element:
                element = element['_nodeid']
                targetcl = stagecluster if stage != 'init' else scopecluster
                shape = 'diamond' if stage == 'init' else 'box'
                label = '' if stage == 'init' else '{}[{}]'.format(stage,i)
                # if not stage=='init':
                #     label += '\n'+', '.join(get_input_par_jsonpaths(workflow.dag.getNode(element).task.attributes))
                additional = {'fixedsize':True, 'height':0.2, 'width':0.2} if stage == 'init' else {}
                targetcl.add_node(pydotplus.graphviz.Node(element, label = label, color = 'blue', shape = shape,**additional))
                add_result(targetcl,element,workflow.dag.getNode(element).result)
            elif type(element)==dict:
                fillscope(stagecluster,workflow,jsonpointer.JsonPointer.from_parts(scopeptr.parts+[stage,i]).path)


def path_to_id(stepid,path):
    return '{}_{}'.format(stepid,path.replace('/','_'))

def get_input_par_jsonpaths(inputpars):
    allleafs = jq.jq('leaf_paths').transform(inputpars, multiple_output=True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x) for x in allleafs]
    return [p.path for p in leafpointers]


def add_result(graph,parent,jsondata):
    allleafs = jq.jq('leaf_paths').transform(jsondata, multiple_output=True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x) for x in allleafs]

    for leaf in leafpointers:
        leafid = path_to_id(parent,leaf.path)
        # value = leaf.resolve(jsondata)
        source = ''.join(['[{}]'.format(p) for p in leaf.parts])
        label = '{}'.format(source)
        graph.add_node(pydotplus.graphviz.Node(leafid, label = label, color = 'red'))
        graph.add_edge(pydotplus.graphviz.Edge(parent,leafid))

def attach_to_results(provgraph,workflow,node):
    for dep in workflow.dag.getNode(node).task.inputs:
        resultid =  path_to_id(dep.stepid,dep.pointer.path)
        provgraph.add_edge(pydotplus.graphviz.Edge(resultid,node))


def connect(provgraph,workflow):
    for node in workflow.dag.nodes():
        attach_to_results(provgraph,workflow,node)

def write_prov_graph(workdir,workflow,vizformat = 'pdf'):
    provgraph = pydotplus.graphviz.Graph()

    fillscope(provgraph,workflow)
    connect(provgraph,workflow)

    with open('{}/yadage_workflow_instance.dot'.format(workdir),'w') as dotfile:
        dotfile.write(provgraph.to_string())

    with open('{}/yadage_workflow_instance.{}'.format(workdir,vizformat),'w') as outfile:
        subprocess.call(shlex.split('dot -T{} {}/yadage_workflow_instance.dot'.format(vizformat,workdir)), stdout = outfile)



#!/usr/bin/env python
import os
import json
import click
import time
import adage
import adage.visualize as av
import packtivity.statecontexts.poxisfs_context as statecontext
import yadage.backends.packtivity_celery
import yadage.backends.celeryapp
import yadage.workflow_loader
import yadage.yadagemodels
import yadage.visualize
import yadage.interactive
import clihelpers
import logging
import serialize
import reset as yr


log = logging.getLogger(__name__)

@click.group()
def mancli():
    pass

@mancli.command()
@click.option('-s','--statefile', default = 'yadage_wflow_state.json')
@click.option('-b','--backendfile', default = 'yadage_backend_state.json')
@click.option('-t','--toplevel', default = os.getcwd())
@click.option('-a','--inputarchive', default = None)
@click.option('--parameter', '-p', multiple=True)
@click.argument('workdir')
@click.argument('workflow')
@click.argument('initfiles', default = '')
def init(workdir,workflow,initfiles,statefile,backendfile,toplevel,parameter,inputarchive):
    workflow_def = yadage.workflow_loader.workflow(
        toplevel = toplevel,
        source = workflow
    )

    rootcontext = statecontext.make_new_context(workdir)
    workflow = yadage.yadagemodels.YadageWorkflow.createFromJSON(workflow_def,rootcontext)

    initdata = clihelpers.getinit_data(initfiles,parameter)
    workflow.view().init(initdata)

    click.secho('initialized workflow', fg = 'green')

    yadagedir = get_yadagedir(workdir)
    '{}/_yadage'.format(workdir)
    os.makedirs(yadagedir)


    statefile = '{}/{}'.format(get_yadagedir(workdir),statefile)
    backendfile = '{}/{}'.format(get_yadagedir(workdir),backendfile)


    click.secho('statefile at {}'.format(statefile))
    serialize.snapshot(
        workflow,
        statefile,
        backendfile
    )

def decide_rule(rule,state):
    click.secho('we could extend DAG with rule', fg = 'blue')
    click.secho('rule: {}/{} ({})'.format(rule.offset,rule.rule.name,rule.identifier))
    shall = raw_input(click.style("Shall we? (y/N) ", fg = 'blue')).lower() == 'y'
    if shall:
        click.secho('ok we will extend.', fg = 'green')
    else:
        click.secho('maybe another time...', fg = 'yellow')
    return shall

def decide_step(dag,nodeobj):
    print 'we could submit a DAG node (id: {}) DAG is: {}'.format(nodeobj,dag)
    shall = raw_input(click.style("Shall we? (y/N) ", fg = 'magenta')).lower() == 'y'
    if shall:
        click.secho('ok we will submit.', fg = 'green')
    else:
        click.secho('will not submit for now...', fg = 'yellow')
    return shall

def custom_decider(decide_func):
    # we yield until we receive some data via send()
    def decider():
        data = yield
        while True:
            data = yield decide_func(*data)
    return decider

def get_yadagedir(workdir):
    return '{}/_yadage'.format(workdir)

def finalize_manual(workdir,workflow):
    click.secho('workflow done.', fg = 'green')
    yadage.visualize.write_prov_graph(get_yadagedir(workdir),workflow)

def load_state(statefile):
    backend = yadage.backends.packtivity_celery.PacktivityCeleryBackend(
        yadage.backends.celeryapp.app
    )

    click.secho('loading state from {}'.format(statefile))
    workflow = yadage.yadagemodels.YadageWorkflow.fromJSON(
        json.load(open(statefile)),
        yadage.backends.packtivity_celery.PacktivityCeleryProxy,
        backend
    )
    return backend,workflow

@mancli.command()
@click.argument('workdir')
@click.option('-s','--statefile', default = 'yadage_wflow_state.json')
@click.option('-b','--backendfile', default = 'yadage_backend_state.json')
@click.option('-v','--verbosity', default = 'ERROR')
def step(workdir,statefile,backendfile,verbosity):
    logging.basicConfig(level = getattr(logging,verbosity))
    statefile = '{}/{}'.format(get_yadagedir(workdir),statefile)
    backendfile = '{}/{}'.format(get_yadagedir(workdir),backendfile)

    backend, workflow = load_state(statefile)

    extend_decider,submit_decider = yadage.interactive.interactive_deciders()
    coroutine = adage.adage_coroutine(backend,extend_decider,submit_decider)
    coroutine.next() #prime the coroutine....

    coroutine.send(workflow)
    try:
        click.secho('try stepping workflow')
        coroutine.next()
    except StopIteration:
        finalize_manual(workdir,workflow)

    serialize.snapshot(
        workflow,
        statefile,
        backendfile
    )
    av.save_dot(av.colorize_graph_at_time(workflow.dag,time.time()).to_string(),'{}/{}'.format(get_yadagedir(workdir),'adage.png'),'png')

@mancli.command()
@click.argument('workdir')
@click.argument('name')
@click.option('-s','--statefile', default = 'yadage_wflow_state.json')
@click.option('-b','--backendfile', default = 'yadage_backend_state.json')
@click.option('-o','--offset', default = '')
def reset(workdir,statefile,backendfile,offset,name):
    statefile = '{}/{}'.format(get_yadagedir(workdir),statefile)
    backendfile = '{}/{}'.format(get_yadagedir(workdir),backendfile)

    backend, workflow = load_state(statefile)
    yr.reset_state(workflow,offset,name)
    serialize.snapshot(
        workflow,
        statefile,
        backendfile
    )
    av.save_dot(av.colorize_graph_at_time(workflow.dag,time.time()).to_string(),'{}/{}'.format(get_yadagedir(workdir),'adage.png'),'png')

if __name__ == '__main__':
    mancli()



import yaml
import json
import click
import tempfile
import shutil
import yadagemodels
import visualize
import backends.staticbackend
from yadagestep import outputReference
from handlers.expression_handlers import handlers as exh
from helpers import set_static_backend

def process(x,dag):
    if type(x)==dict:
        for k,v in x.iteritems():
            x[k] = process(v,dag)
        return x
    elif type(x)==list:
        for i,e in enumerate(x):
            x[i] = process(e,dag)
        return x
    elif type(x) == outputReference:
        return x.pointer.resolve(dag.getNode(x.stepid).result)
    else:
        return x

def printRef(ref,dag,indent = ''):
    click.secho('{}name: {} position: {}, value: {}, id: {}'.format(
        indent,
        dag.getNode(ref.stepid).name,
        ref.pointer.path,
        ref.pointer.resolve(dag.getNode(ref.stepid).result),
        ref.stepid
        ),
    fg = 'cyan')

@click.group()
def utilcli():
    pass

@utilcli.command()
@click.argument('instance')
@click.argument('results')
@click.argument('selection')
def testsel(instance,results,selection):
    wflow = yadagemodels.YadageWorkflow.fromJSON(
        json.load(open(instance))
    )
    backend = backends.staticbackend.StaticBackend(json.load(open(results)))
    set_static_backend(wflow.dag,backend)

    selresult = exh['stage-output-selector'](wflow.view(),yaml.load(selection))
    if not selresult:
        click.secho('Bad selection', fg = 'red')
        return


    click.secho(json.dumps(
            process(selresult,wflow.dag),
            sort_keys=True,
            indent=4,
            separators=(',', ': ')),
        fg = 'green'
    )

@utilcli.command()
@click.argument('instance')
@click.argument('results')
@click.argument('vizpdf')
def viz(instance,results,vizpdf):
    wflow = yadagemodels.YadageWorkflow.fromJSON(
        json.load(open(instance))
    )
    backend = backends.staticbackend.StaticBackend(json.load(open(results)))
    set_static_backend(wflow.dag,backend)

    dirpath = tempfile.mkdtemp()
    visualize.write_prov_graph(dirpath,wflow)
    shutil.copy('{}/yadage_workflow_instance.pdf'.format(dirpath),vizpdf)
    shutil.rmtree(dirpath)


if __name__ == '__main__':
    utilcli()



import json
import os
import subprocess

pwd = os.path.abspath(os.curdir)
basecmd = 'docker run -e KRBUSER={krbuser} -v {keytab}:/auth/krb.keytab -v {pwd}:{pwd} -w {pwd} --rm -i cernjiracurl'.format(
    krbuser = 'lheinric', keytab = '/Users/lukas/lheinric.keytab', pwd = pwd
)

def create(summary, description, parent = None):
    createtask = {
        "fields": {
           "issuetype": {"name": "Task"},
           "project": {"key": "RCST"},
           "summary": summary,
           "description": description,
       }
    }
    if parent:
        createtask['fields']['issuetype'] = {"name":"Sub-task"}
        createtask['fields']['parent'] = {"key":parent}

    p = subprocess.Popen(
        '{base} {cmd}'.format(base = basecmd, cmd = 'create.sh'),
        shell = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE
    )
    out, err = p.communicate(json.dumps(createtask))
    response = json.loads(out)
    print response
    return response

def attach(data,task,remotename):
    p = subprocess.Popen(
        '{base} {cmd}'.format(base = basecmd, cmd = 'attach.sh {} {}'.format(task,remotename)),
        shell = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE
    )
    out, err = p.communicate(json.dumps(dict(data)))
    response = json.loads(out)
    return response

def getissue(task):
    p = subprocess.Popen(
        '{base} {cmd}'.format(base = basecmd, cmd = 'getissue.sh {}'.format(task)),
        shell = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE
    )
    out, err = p.communicate()
    response = json.loads(out)
    return response

def resolve(task,sucess,publishdata = None):
    cmd = 'done.sh' if sucess else 'wontdo.sh'
    p = subprocess.Popen(
        '{base} {cmd}'.format(base = basecmd, cmd = '{} {}'.format(cmd,task)),
        shell = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE
    )
    out, err = p.communicate()
    if publishdata:
        attach(publishdata,task,'published.json')

def getfile(task,filename):
    taskdata = getissue(task)
    attdata = {x['filename']:x['content'] for x in taskdata['fields']['attachment']}
    url = attdata[filename]
    p = subprocess.Popen(
        '{base} {cmd}'.format(base = basecmd, cmd = 'getatt.sh {}'.format(url)),
        shell = True, stdin = subprocess.PIPE, stdout = subprocess.PIPE
    )
    out, err = p.communicate()
    response = json.loads(out)
    return response

def getstatus(task):
    data = getissue(task)
    return  data['fields']['resolution']

def submit(name,spec,parameters,context,parent):
    data = create('manual packtivity {}'.format(name),'this is a description',parent)
    taskid = data['key']
    attach(parameters,taskid,'parameters.json')
    attach(spec,taskid,'spec.json')
    attach(context,taskid,'context.json')
    taskdata = getissue(taskid)
    return taskid


# workflowtask = create('workflow request #{}'.format(1234),'this is a description')
# taskid =  submit('first',{'this':'is a first spec'},{'and':'two','param':'values'},{'a first':'context'},workflowtask['key'])
# taskid =  submit('second',{'this':'is a another spec'},{'and':'two','param':'values'},{'another':'context'},workflowtask['key'])

# resolve(taskid,True,{'a':'response'})
# print getfile(taskid,'published.json')



#!/usr/bin/env python
import adage
import adage.backends
import logging
import os
import json
import workflow_loader
from yadage.yadagemodels import YadageWorkflow
import visualize
import serialize
import interactive
import  packtivity.statecontexts.poxisfs_context as statecontext


log = logging.getLogger(__name__)

RC_FAILED = 1
RC_SUCCEEDED = 0

def run_workflow(
    workdir,
    workflow,
    initdata,
    loadtoplevel,
    updateinterval,
    loginterval,
    schemadir,
    backend,
    user_interaction = False,
    validate = True,
    doviz = True
    ):
    """
    Main entry point to run a Yadage workflow
    """
    #let's be conservative and just assume we're going to fail. will set success RC explicityly
    return_value = RC_FAILED

    log.info('running yadage workflow %s',workflow)
    rootcontext = statecontext.make_new_context(workdir)

    for k,v in initdata.iteritems():
        candpath = '{}/inputs/{}'.format(workdir,v)
        if os.path.exists(candpath):
            initdata[k] = '{}/inputs/{}'.format(rootcontext['readwrite'][0],v)

    workflow_json = workflow_loader.workflow(
        workflow,
        toplevel = loadtoplevel,
        schemadir = schemadir,
        validate = validate
    )
    workflow = YadageWorkflow.createFromJSON(workflow_json,rootcontext)
    workflow.view().init(initdata)

    yadagedir = '{}/_yadage/'.format(workdir)
    os.makedirs(yadagedir)

    with open('{}/yadage_instance_before.json'.format(yadagedir),'w') as f:
        json.dump(workflow.json(),f)

    if user_interaction:
        extend,submit = interactive.interactive_deciders()
        interactive_kwargs = {
            'extend_decider': extend,
            'submit_decider': submit
        }
    else:
        interactive_kwargs = {}

    try:
        adage.rundag(workflow,
                     default_trackers = doviz,
                     backend = backend,
                     update_interval = updateinterval,
                     trackevery = loginterval,
                     workdir = '{}/_adage'.format(workdir),
                     **interactive_kwargs
        )
        return_value = RC_SUCCEEDED
    except:
        log.exception('Unfortunately we failed. :(')

    serialize.snapshot(
        workflow,
        '{}/yadage_snapshot_workflow.json'.format(yadagedir),
        '{}/yadage_snapshot_backend.json'.format(yadagedir)
    )

    with open('{}/yadage_template.json'.format(yadagedir),'w') as f:
        json.dump(workflow_json,f)
    
    if doviz:
        visualize.write_prov_graph(yadagedir,workflow, vizformat = 'png')
        visualize.write_prov_graph(yadagedir,workflow, vizformat = 'pdf')
    log.info('finished yadage workflow %s, returning rc: %s',workflow,return_value)
    return return_value



import json
def snapshot(workflow,wflowfile,resultfile):
    '''
    snapshots the state of the workflow into two files:
        - a file holding the pure state of the workflow
        - a file holding the backend state, i.e. results to specific tasks
    '''
    with open(wflowfile,'w') as wfile:
        json.dump(workflow.json(),wfile)

    with open(resultfile,'w') as rfile:
        json.dump({k:workflow.dag.getNode(k).result for k in workflow.dag.nodes()},rfile)



import logging
import utils
import jq
import jsonpointer
import copy
from yadage.yadagestep import outputReference
import jsonpath_rw

log = logging.getLogger(__name__)

handlers,scheduler = utils.handler_decorator()

def pointerize(jsondata, asref=False, stepid=None):
    '''
    a helper method that replaces leaf nodes in a JSON object with
    a outputReference objects (~ a JSONPath) pointing to that leaf position
    useful to track access to leaf nodes later on.
    '''
    allleafs = jq.jq('leaf_paths').transform(jsondata, multiple_output=True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x).path for x in allleafs]
    jsondata_proxy = copy.deepcopy(jsondata)
    for leaf in leafpointers:
        x = jsonpointer.JsonPointer(leaf)
        x.set(jsondata_proxy, outputReference(stepid, x) if asref else x.path)
    return jsondata_proxy


def select_reference(step, selection):
    '''
    resolves a jsonpath selection and returns JSONPointerized matches
    '''
    log.debug('selecting output from step %s',step)
    pointerized = pointerize(step['result'], asref=True, stepid=step['id'])
    matches = jsonpath_rw.parse(selection).find(pointerized)
    if not matches:
        log.error('no matches found for selection %s in result %s', selection, step['result'])
        raise RuntimeError('no matches found in reference selection')

    if len(matches) > 1:
        log.error('found multiple matches to query: %s within result: %s\n \ matches %s', selection, step['result'], matches)
        raise RuntimeError('multiple matches in result jsonpath query')
    return matches[0].value


def combine_outputs(outputs, flatten, unwrapsingle):
    '''
    combines the result of multiple reference selections into a single outputs.
    non-list values will just be combined into a list while lists will be concatenated
    optinally we can return the sole element of a single-value list (argument: unwrapsingle)
    '''
    combined = []
    for reference in outputs:
        if type(reference)==list:
            if flatten:
                for elementref in reference:
                    combined+=[elementref]
            else:
                combined+=[reference]
        else:
            combined+=[reference]
    if len(combined)==1 and unwrapsingle:
        combined = combined[0]
    return combined

def select_steps(stageview,query):
    '''
    selects the step objects from the stage view based on a query and converts them to simple
    dictionaries with id,result keys
    '''
    return [{'id':s.identifier, 'result':s.result} for s in stageview.getSteps(query)]

def select_outputs(steps,selection,flatten,unwrapsingle):
    return combine_outputs(map(lambda s: select_reference(s, selection), steps), flatten, unwrapsingle)

@scheduler('stage-output-selector')
def resolve_reference(stageview,selection):
    '''resolves a output reference by selecting the stage and stage outputs'''
    log.debug('resolving selection %s',selection)
    if type(selection) is not dict:
        return None
    else:
        steps = select_steps(stageview, selection['stages'])
        log.debug('selected steps %s',steps)
        outputs = select_outputs(steps,
                                 selection['output'],
                                 selection.get('flatten', False),
                                 selection.get('unwrap', False))
        log.debug('selected outputs %s',outputs)
        return outputs



import logging
import utils
import itertools

from expression_handlers import handlers as exprhandlers
from yadage.yadagestep import yadagestep, initstep, outputReference
from yadage.yadagemodels import jsonStage
import  packtivity.statecontexts.poxisfs_context as statecontext

log = logging.getLogger(__name__)

handlers, scheduler = utils.handler_decorator()

# A scheduler does the following things:
#   - attached new nodes to the DAG
# - for each added step
#     - the step is given a name
#     - the step attributes are determined using the scheduler spec and context
#     - a list of used inputs (in the form of [stepname,outputkey,index])

def select_parameter(stageview,parameter):
    if type(parameter) is not dict:
        value = parameter
    else:
        handler = exprhandlers[parameter['expression_type']]
        value = handler(stageview,parameter)
    return value

def finalize_value(stage,step,value,context):
    '''
    finalize a value by recursively resolving references and
    interpolating with the context when necessary
    '''
    if type(value)==outputReference:
        step.used_input(value)
        v = value.pointer.resolve(stage.view.dag.getNode(value.stepid).result)
        return finalize_value(stage,step,v,context)
    if type(value)==list:
        return [finalize_value(stage,step,x,context) for x in value]
    if type(value) in [str,unicode]:
        return value.format(**context)
    return value

def finalize_input(stage,step,json,context):
    '''
    evaluate final values of parameters by either resolving a
    reference to a upstream output or evaluating a static
    reference from the template (possibly string-interpolated)
    '''

    context = context.copy()
    context['workdir'] = context['readwrite'][0]
    result = {}
    for k,v in json.iteritems():
        if type(v) is not list:
            result[k] = finalize_value(stage,step,v,context)
        else:
            result[k] = [finalize_value(stage,step,element,context) for element in v]

    return result

def step_or_init(name,spec,context):
    if 'step' in spec:
        stepcontext = statecontext.make_new_context(name,context, subdir = True)
        return yadagestep(name = name, spec = spec['step'], context = stepcontext)
    elif 'workflow' in spec:
        return initstep('init {}'.format(name))


def addStepOrWorkflow(name,stage,step,spec):
    if type(step)==initstep:
        newcontext = statecontext.make_new_context(name,stage.context)
        subrules = [jsonStage(yml,newcontext) for yml in spec['workflow']['stages']]
        stage.addWorkflow(subrules, initstep = step)
    else:
        stage.addStep(step)

def get_parameters(spec):
    return {x['key']:x['value']for x in spec['parameters']}

@scheduler('singlestep-stage')
def singlestep_stage(stage,spec):
    '''
    a simple state that adds a single step/workflow. The node is attached
    to the DAG based on used upstream outputs
    '''
    log.debug('scheduling singlestep stage with spec:\n%s',spec)

    step = step_or_init(name = stage.name, spec = spec, context = stage.context)
    ctx = step.context if hasattr(step,'context') else stage.context

    parameters = {
        k:select_parameter(stage.view,v) for k,v in get_parameters(spec).iteritems()
    }
    finalized = finalize_input(stage,step,parameters,ctx)

    addStepOrWorkflow(stage.name,stage,step.s(**finalized),spec)

def scatter(parameters,scatter):
    '''
    this method turns a parameter set and scatter definition into a list
    of single parameter sets.
    '''
    commonpars = parameters.copy()
    to_scatter = {}
    for scatpar in scatter['parameters']:
        to_scatter[scatpar] = commonpars.pop(scatpar)

    singlesteppars=[]
    if scatter['method']=='zip':
        keys, zippable = zip(*[(k,v) for i,(k,v) in enumerate(to_scatter.iteritems())])

        for zipped in zip(*zippable):
            individualpars = dict(zip(keys,zipped))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]

    if scatter['method']=='cartesian':
        for what in itertools.product(*[to_scatter[k] for k in scatter['parameters']]):
            individualpars = dict(zip(scatter['keys'],what))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]
    return singlesteppars

@scheduler('multistep-stage')
def multistep_stage(stage,spec):
    '''
    a stage that attaches an array of nodes to the DAG. The number of nodes
    is determined by a scattering recipe. Currently two algs are supported
    'zip': one or more arrays of length n are iterated through in lock-step.
           n nodes are added to the DAG where the  parameters values are set to
           the values in the iteration
    'cartesian': a cartesian product of a number of arrays (possibly different sizes)
                 adds n1 x n2 x ... nj nodes.
    Nodes are attached to the DAG based on used upstream inputs
    '''
    log.debug('scheduling multistep stage with spec:\n%s',spec)
    parameters = {
        k:select_parameter(stage.view,v) for k,v in get_parameters(spec).iteritems()
    }
    singlesteppars = scatter(parameters,spec['scatter'])
    for i,pars in enumerate(singlesteppars):
        singlename = '{}_{}'.format(stage.name,i)
        step = step_or_init(name = singlename, spec = spec, context = stage.context)
        ctx = step.context if hasattr(step,'context') else stage.context
        ctx = ctx.copy()
        ctx.update(index = i)
        finalized = finalize_input(stage,step,pars,ctx)
        addStepOrWorkflow(singlename,stage,step.s(**finalized),spec)



class StaticProxy(object):
    def __init__(self,taskid):
        self.taskid = taskid

class StaticBackend(object):
    def __init__(self,resultdata):
        self.resultdata = resultdata

    def submit(self,task):
        raise RuntimeError('this backend is not meant for submission but just for loading a workflow')

    def result(self,resultproxy):
        return self.resultdata[resultproxy.taskid]

    def ready(self,resultproxy):
        return True

    def successful(self,resultproxy):
        raise NotImplementedError

    def fail_info(self,resultproxy):
        return 'cannot give reason :('



import jirainterface
from packtivitybackend import PacktivityProxyBase
import yadage.yadagestep

class JiraProxy(PacktivityProxyBase):
    def proxyname(self):
        return 'JiraPacktivityProxy'

    @classmethod
    def fromJSON(cls,data):
        proxy = data['proxydetails']
        return cls(proxy)


class JiraBackend(object):
    def __init__(self,title = None, description = None, wflowid = None):
        if wflowid:
            self.parenttask = wflowid
        else:
            data = jirainterface.create('workflow request: {}'.format(title),description)
            self.parenttask = data['key']

    def submit(self,task):
        tasktype = type(task)
        if tasktype == yadage.yadagestep.yadagestep:
            spec = task.spec
            context = task.context
        else:
            spec = {'no':'spec'}
            context = {'no':'context'}

        taskid = jirainterface.submit(task.name,spec,task.attributes,context,self.parenttask)
        return JiraProxy({'task_id':taskid, 'task_name':task.name})

    def result(self,resultproxy):
        return jirainterface.getfile(resultproxy.proxy['task_id'],'published.json')

    def ready(self,resultproxy):
        status = jirainterface.getstatus(resultproxy.proxy['task_id'])
        return True if status else False

    def successful(self,resultproxy):
        status = jirainterface.getstatus(resultproxy.proxy['task_id'])
        return status['name'] == 'Done'

    def fail_info(self,resultproxy):
        return 'cannot give reason :('
